use std::{
    collections::{HashMap, HashSet, VecDeque},
    fmt::Pointer,
    pin::pin,
    sync::{atomic::AtomicBool, Arc},
    task::{ready, Poll},
};

use futures::{
    stream::{SplitSink, SplitStream},
    Sink, SinkExt, Stream, StreamExt,
};
use jabber::{
    connection::Tls,
    jabber_stream::bound_stream::{BoundJabberReader, BoundJabberStream, BoundJabberWriter},
    JID,
};
use sqlx::{query, Pool, Sqlite, SqlitePool};
use stanza::{
    client::{
        iq::{Iq, IqType, Query},
        Stanza,
    },
    roster,
};
use tokio::{
    io::AsyncRead,
    select,
    sync::{
        mpsc::{self, Receiver, Sender},
        oneshot, Mutex,
    },
    task::{JoinHandle, JoinSet},
};
use tokio_stream::wrappers::ReceiverStream;
use tokio_util::sync::{PollSendError, PollSender};

// actor that receives jabber stanzas to write, and if there is an error, sends a message back to the supervisor then aborts, so the supervisor can spawn a new stream.
pub struct JabberWriter {
    stanza_receiver: mpsc::Receiver<JabberWrite>,
    control_receiver: mpsc::Receiver<JabberWriterControl>,
    stream: BoundJabberWriter<Tls>,
    on_crash: oneshot::Sender<(JabberWrite, mpsc::Receiver<JabberWrite>)>,
}

struct JabberWrite {
    stanza: Stanza,
    respond_to: oneshot::Sender<Result<(), Error>>,
}

enum JabberWriterControl {
    Disconnect,
    Abort(oneshot::Sender<mpsc::Receiver<JabberWrite>>),
}

impl JabberWriter {
    fn new(
        stanza_receiver: mpsc::Receiver<JabberWrite>,
        control_receiver: mpsc::Receiver<JabberWriterControl>,
        stream: BoundJabberWriter<Tls>,
        supervisor: oneshot::Sender<(JabberWrite, mpsc::Receiver<JabberWrite>)>,
    ) -> Self {
        Self {
            stanza_receiver,
            control_receiver,
            stream,
            on_crash: supervisor,
        }
    }

    async fn write(&mut self, stanza: &Stanza) -> Result<(), peanuts::Error> {
        Ok(self.stream.write(stanza).await?)
    }

    async fn run(mut self) {
        loop {
            tokio::select! {
                Some(msg) = self.control_receiver.recv() => {
                    match msg {
                        JabberWriterControl::Disconnect => {
                            // TODO: close the stanza_receiver channel and drain out all of the remaining stanzas to send
                            self.stanza_receiver.close();
                            // TODO: put this in some kind of function to avoid code duplication
                            for msg in self.stanza_receiver.recv().await {
                                let result = self.stream.write(&msg.stanza).await;
                                match result {
                                    Err(e) => match &e {
                                        peanuts::Error::ReadError(error) => {
                                            // make sure message is not lost from error, supervisor handles retry and reporting
                                            self.on_crash.send((msg, self.stanza_receiver));
                                            break;
                                        }
                                        _ => {
                                            msg.respond_to.send(Err(e.into()));
                                        }
                                    },
                                    _ => {
                                        msg.respond_to.send(Ok(()));
                                    }
                                }
                            }
                            self.stream.try_close().await;
                            break;
                        },
                        JabberWriterControl::Abort(sender) => {
                            sender.send(self.stanza_receiver);
                            break;
                        },
                    }
                },
                Some(msg) = self.stanza_receiver.recv() => {
                    let result = self.stream.write(&msg.stanza).await;
                    match result {
                        Err(e) => match &e {
                            peanuts::Error::ReadError(error) => {
                                // make sure message is not lost from error, supervisor handles retry and reporting
                                self.on_crash.send((msg, self.stanza_receiver));
                                break;
                            }
                            _ => {
                                msg.respond_to.send(Err(e.into()));
                            }
                        },
                        _ => {
                            msg.respond_to.send(Ok(()));
                        }
                    }
                },
                // TODO: check if this is ok to do
                else => break,
            }
        }
    }
}

#[derive(Clone)]
pub struct JabberWriteHandle {
    sender: mpsc::Sender<JabberWrite>,
}

pub struct JabberWriterControlHandle {
    sender: mpsc::Sender<JabberWriterControl>,
    handle: JoinHandle<()>,
}

impl JabberWriterControlHandle {
    pub fn new(
        stream: BoundJabberWriter<Tls>,
        supervisor: oneshot::Sender<(JabberWrite, mpsc::Receiver<JabberWrite>)>,
    ) -> (JabberWriteHandle, JabberWriterControlHandle) {
        let (control_sender, control_receiver) = mpsc::channel(20);
        let (stanza_sender, stanza_receiver) = mpsc::channel(20);

        let actor = JabberWriter::new(stanza_receiver, control_receiver, stream, supervisor);
        let handle = tokio::spawn(async move { actor.run().await });

        (
            JabberWriteHandle {
                sender: stanza_sender,
            },
            Self {
                sender: control_sender,
                handle,
            },
        )
    }

    pub fn reconnect(
        stream: BoundJabberWriter<Tls>,
        supervisor: oneshot::Sender<(JabberWrite, mpsc::Receiver<JabberWrite>)>,
        stanza_receiver: mpsc::Receiver<JabberWrite>,
    ) -> Self {
        let (control_sender, control_receiver) = mpsc::channel(20);

        let actor = JabberWriter::new(stanza_receiver, control_receiver, stream, supervisor);
        let handle = tokio::spawn(async move { actor.run().await });

        Self {
            sender: control_sender,
            handle,
        }
    }
}

pub struct JabberReader {
    // TODO: place iq hashmap here
    control_receiver: mpsc::Receiver<JabberReaderControl>,
    stream: BoundJabberReader<Tls>,
    on_crash: oneshot::Sender<(SqlitePool, mpsc::Sender<UpdateMessage>, JoinSet<()>)>,
    db: SqlitePool,
    sender: mpsc::Sender<UpdateMessage>,
    supervisor_control: mpsc::Sender<JabberSupervisorCommand>,
    write_sender: mpsc::Sender<JabberWrite>,
    tasks: JoinSet<()>,
}

impl JabberReader {
    fn new(
        control_receiver: mpsc::Receiver<JabberReaderControl>,
        stream: BoundJabberReader<Tls>,
        on_crash: oneshot::Sender<(SqlitePool, mpsc::Sender<UpdateMessage>, JoinSet<()>)>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
        // jabber server must be able to both terminate the connection from error, and ask for data from the client (such as supported XEPs)
        supervisor_control: mpsc::Sender<JabberSupervisorCommand>,
        write_sender: mpsc::Sender<JabberWrite>,
    ) -> Self {
        Self {
            control_receiver,
            stream,
            on_crash,
            db,
            sender,
            supervisor_control,
            write_sender,
            tasks: JoinSet::new(),
        }
    }

    async fn run(mut self) {
        loop {
            tokio::select! {
                Some(msg) = self.control_receiver.recv() => {
                    match msg {
                        JabberReaderControl::Disconnect => todo!(),
                        JabberReaderControl::Abort(sender) => todo!(),
                    };
                },
                stanza = self.stream.read::<Stanza>() => {
                    match stanza {
                        Ok(_) => todo!(),
                        Err(_) => todo!(),
                    }
                    self.tasks.spawn();
                },
                else => break
            }
        }
    }
}

trait Task {
    async fn handle();
}

impl Task for Stanza {
    async fn handle() {
        todo!()
    }
}

enum JabberReaderControl {
    Disconnect,
    Abort(oneshot::Sender<mpsc::Receiver<JabberWrite>>),
}

struct JabberReaderControlHandle {
    sender: mpsc::Sender<JabberReaderControl>,
    handle: JoinHandle<()>,
}

impl JabberReaderControlHandle {
    pub fn new(
        stream: BoundJabberReader<Tls>,
        on_crash: oneshot::Sender<(SqlitePool, mpsc::Sender<UpdateMessage>, JoinSet<()>)>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
        supervisor_control: mpsc::Sender<JabberSupervisorCommand>,
        jabber_write: mpsc::Sender<JabberWrite>,
    ) -> Self {
        let (control_sender, control_receiver) = mpsc::channel(20);

        let actor = JabberReader::new(
            control_receiver,
            stream,
            on_crash,
            db,
            sender,
            supervisor_control,
            jabber_write,
        );
        let handle = tokio::spawn(async move { actor.run().await });

        Self {
            sender: control_sender,
            handle,
        }
    }
}

pub struct Luz {
    receiver: mpsc::Receiver<CommandMessage>,
    jid: Arc<Mutex<JID>>,
    // TODO: use a dyn passwordprovider trait to avoid storing password in memory
    password: String,
    connected: Arc<Mutex<Option<(JabberWriteHandle, JabberSupervisorHandle)>>>,
    db: SqlitePool,
    sender: mpsc::Sender<UpdateMessage>,
    tasks: JoinSet<()>,
}

impl Luz {
    fn new(
        receiver: mpsc::Receiver<CommandMessage>,
        jid: Arc<Mutex<JID>>,
        password: String,
        connected: Arc<Mutex<Option<(JabberWriteHandle, JabberSupervisorHandle)>>>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
    ) -> Self {
        Self {
            jid,
            password,
            connected,
            db,
            receiver,
            sender,
            tasks: JoinSet::new(),
        }
    }

    async fn run(mut self) {
        while let Some(msg) = self.receiver.recv().await {
            // TODO: consider separating disconnect/connect and commands apart from commandmessage
            match msg {
                CommandMessage::Connect => {
                    match self.connected.lock().await.as_ref() {
                        Some(_) => {
                            self.sender
                                .send(UpdateMessage::Error(Error::AlreadyConnected))
                                .await;
                        }
                        None => {
                            let mut jid = self.jid.lock().await;
                            let mut domain = jid.domainpart.clone();
                            let streams_result =
                                jabber::connect_and_login(&mut jid, &self.password, &mut domain)
                                    .await;
                            match streams_result {
                                Ok(s) => {
                                    let (writer, supervisor) =
                                        JabberSupervisorHandle::new(s, self.sender.clone());
                                    *self.connected.lock().await = Some((writer, supervisor));
                                }
                                Err(e) => {
                                    self.sender.send(UpdateMessage::Error(e.into()));
                                }
                            }
                        }
                    };
                }
                CommandMessage::Disconnect => match self.connected.lock().await.as_mut() {
                    None => {
                        self.sender
                            .send(UpdateMessage::Error(Error::AlreadyDisonnected))
                            .await;
                    }
                    mut c => {
                        if let Some((_write_handle, supervisor_handle)) = c.take() {
                            let _ = supervisor_handle
                                .sender
                                .send(JabberSupervisorCommand::Disconnect)
                                .await;
                        } else {
                            unreachable!()
                        };
                    }
                },
                _ => {
                    match self.connected.lock().await.as_ref() {
                        Some((w, _)) => self.tasks.spawn(msg.handle_online(
                            w.clone(),
                            self.jid.clone(),
                            self.db.clone(),
                            self.sender.clone(),
                        )),
                        None => self.tasks.spawn(msg.handle_offline(
                            self.jid.clone(),
                            self.db.clone(),
                            self.sender.clone(),
                        )),
                    };
                }
            }
        }
    }
}

impl CommandMessage {
    pub async fn handle_offline(
        mut self,
        jid: Arc<Mutex<JID>>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
    ) {
        todo!()
    }

    pub async fn handle_online(
        mut self,
        jabber_write_handle: JabberWriteHandle,
        // TODO: jid could lose resource by the end
        jid: Arc<Mutex<JID>>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
    ) {
        todo!()
    }
}

// TODO: separate sender and receiver, store handle to Luz process to ensure dropping
pub struct LuzHandle {
    sender: mpsc::Sender<CommandMessage>,
    receiver: mpsc::Receiver<UpdateMessage>,
}

impl LuzHandle {
    pub fn new(jid: JID, password: String, db: SqlitePool) -> Self {
        let (command_sender, command_receiver) = mpsc::channel(20);
        let (update_sender, update_receiver) = mpsc::channel(20);

        let actor = Luz::new(
            command_receiver,
            Arc::new(Mutex::new(jid)),
            password,
            Arc::new(Mutex::new(None)),
            db,
            update_sender,
        );
        tokio::spawn(async move { actor.run().await });

        Self {
            sender: command_sender,
            receiver: update_receiver,
        }
    }
}

pub struct JabberSupervisor {
    connection_commands: mpsc::Receiver<JabberSupervisorCommand>,
    writer_crash: oneshot::Receiver<(JabberWrite, mpsc::Receiver<JabberWrite>)>,
    reader_crash: oneshot::Receiver<(SqlitePool, mpsc::Sender<UpdateMessage>, JoinSet<()>)>,
    sender: mpsc::Sender<UpdateMessage>,
    writer_handle: JabberWriterControlHandle,
    reader_handle: JabberReaderControlHandle,
}

pub enum JabberSupervisorCommand {
    Disconnect,
}

impl JabberSupervisor {
    fn new(
        connection_commands: mpsc::Receiver<JabberSupervisorCommand>,
        writer_crash: oneshot::Receiver<(JabberWrite, mpsc::Receiver<JabberWrite>)>,
        reader_crash: oneshot::Receiver<(SqlitePool, mpsc::Sender<UpdateMessage>, JoinSet<()>)>,
        sender: mpsc::Sender<UpdateMessage>,
        writer_handle: JabberWriterControlHandle,
        reader_handle: JabberReaderControlHandle,
    ) -> Self {
        Self {
            connection_commands,
            writer_crash,
            sender,
            writer_handle,
            reader_handle,
            reader_crash,
        }
    }

    async fn handle_command_message(&mut self, msg: JabberSupervisorCommand) {}

    async fn run(mut self) {
        loop {
            tokio::select! {
                Some(msg) = self.connection_commands.recv() => {
                    self.handle_command_message(msg).await;
                },
                error = self.writer_crash => {

                },
                error = self.reader_crash => {

                },
            }
        }
    }
}

pub struct JabberSupervisorHandle {
    sender: mpsc::Sender<JabberSupervisorCommand>,
    handle: JoinHandle<()>,
}

impl JabberSupervisorHandle {
    pub fn new(
        streams: BoundJabberStream<Tls>,
        sender: mpsc::Sender<UpdateMessage>,
        db: SqlitePool,
        update_sender: mpsc::Sender<UpdateMessage>,
    ) -> (JabberWriteHandle, Self) {
        let (command_sender, command_receiver) = mpsc::channel(20);
        let (writer_error_sender, writer_error_receiver) = oneshot::channel();
        let (reader_crash_sender, reader_crash_receiver) = oneshot::channel();

        let (reader, writer) = streams.split();
        let (jabber_write_handle, jabber_writer_control_handle) =
            JabberWriterControlHandle::new(writer, writer_error_sender);
        let jabber_reader_control_handle = JabberReaderControlHandle::new(
            reader,
            reader_crash_sender,
            db,
            update_sender,
            command_sender.clone(),
            jabber_write_handle.sender.clone(),
        );

        let actor = JabberSupervisor::new(
            command_receiver,
            writer_error_receiver,
            reader_crash_receiver,
            sender,
            jabber_writer_control_handle,
            jabber_reader_control_handle,
        );

        let handle = tokio::spawn(async move { actor.run().await });

        (
            jabber_write_handle,
            Self {
                sender: command_sender,
                handle,
            },
        )
    }
}

pub enum Error {
    AlreadyConnected,
    PollSend(PollSendError<CommandMessage>),
    Jabber(jabber::Error),
    XML(peanuts::Error),
    SQL(sqlx::Error),
    JID(jid::ParseError),
    AlreadyDisonnected,
}

impl From<peanuts::Error> for Error {
    fn from(e: peanuts::Error) -> Self {
        Self::XML(e)
    }
}

impl From<jid::ParseError> for Error {
    fn from(e: jid::ParseError) -> Self {
        Self::JID(e)
    }
}

impl From<sqlx::Error> for Error {
    fn from(e: sqlx::Error) -> Self {
        Self::SQL(e)
    }
}

impl From<jabber::Error> for Error {
    fn from(e: jabber::Error) -> Self {
        Self::Jabber(e)
    }
}

impl From<PollSendError<CommandMessage>> for Error {
    fn from(e: PollSendError<CommandMessage>) -> Self {
        Self::PollSend(e)
    }
}

pub enum CommandMessage {
    Connect,
    Disconnect,
    /// gets the roster. if offline, retreives cached version from database. should be stored in application memory.
    GetRoster,
    SendMessage(JID, String),
}

pub enum UpdateMessage {
    Error(Error),
    Roster(Vec<roster::Item>),
}
